from hwt.hdl.types.bits import Bits


class SystemCSerializer_type():
    @classmethod
    def HdlType_bits(cls, typ, ctx, declaration=False):
        if declaration:
            raise NotImplementedError()

        w = typ.bit_length()

        if w <= 64:
            if typ.signed:
                typeBaseName = "int"
            else:
                typeBaseName = "uint"
        else:
            if typ.signed:
                typeBaseName = "bigint"
            else:
                typeBaseName = "biguint"

        return "sc_%s<%d>" % (typeBaseName, w)

    @classmethod
    def HdlType_enum(cls, typ, ctx, declaration=False):
        if declaration:
            raise TypeError(
                "There is problem with tracing of c enums, use Bits instead")
        else:
            valueCnt = len(typ._allValues)
            return cls.HdlType_bits(Bits(valueCnt.bit_length()), ctx,
                                    declaration=declaration)



from hwt.interfaces.agents.signal import SignalAgent
from hwt.interfaces.agents.vldSynced import VldSyncedAgent
from hwt.simulator.agentBase import SyncAgentBase, AgentBase


class RegCntrlAgent(SyncAgentBase):
    """
    Simulation/verification agent for RegCntrl interface
    """

    def __init__(self, intf):
        AgentBase.__init__(self, intf)
        super().__init__(intf)

        self._din = SignalAgent(intf.din)
        self._dout = VldSyncedAgent(intf.dout, allowNoReset=True)

    def din_getter(self):
        return self._din.data

    def din_setter(self, newVal):
        self._din.data = newVal

    din = property(din_getter, din_setter)

    def dout_getter(self):
        return self._dout.data

    def dout_setter(self, newVal):
        self._dout.data = newVal

    dout = property(dout_getter, dout_setter)

    def getDrivers(self):
        return self._din.getMonitors() + self._dout.getDrivers()

    def getMonitors(self):
        return self._din.getDrivers() + self._dout.getMonitors()



from hwt.hdl.constants import Time
from hwt.simulator.agentBase import AgentBase
from hwt.simulator.shortcuts import CallbackLoop


class OscilatorAgent(AgentBase):
    """
    Simulation agent for :class:`hwt.interfaces.std.Clk` interface

    * In driver mode oscillates at frequency specified by period

    * In monitor driver captures tuples (time, nextVal) for each change
        on signal (nextVal is 1/0/None)

    :ivar period: period of signal to generate
    :ivar initWait: time to wait before starting oscillation
    """

    def __init__(self, intf, period=10 * Time.ns):
        super(OscilatorAgent, self).__init__(intf)
        self.period = period
        self.initWait = 0
        self.intf = self.intf._sigInside
        self.monitor = CallbackLoop(self.intf, self.monitor, self.getEnable)

    def driver(self, sim):
        sig = self.intf
        sim.write(0, sig)
        halfPeriod = self.period / 2
        yield sim.wait(self.initWait)

        while True:
            yield sim.wait(halfPeriod)
            sim.write(1, sig)
            yield sim.wait(halfPeriod)
            sim.write(0, sig)

    def getMonitors(self):
        self.last = (-1, None)
        self.data = []

        return [self.monitor]

    def monitor(self, sim):
        yield sim.waitOnCombUpdate()
        v = sim.read(self.intf)
        if not v.vldMask:
            v = None
        else:
            v = v.val

        now = sim.now
        last = self.last

        _next = (now, v)
        if last[0] == now:
            if last[1] is not v:
                # update last value
                last = (now, v)
                self.last = last
                if self.data:
                    self.data[-1] = last
                else:
                    self.data.append(last)
            else:
                return
        else:
            self.data.append(_next)

        self.last = _next



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

from setuptools import setup, find_packages


setup(name='hwt',
      version='2.1',
      description='hdl synthesis toolkit',
      url='https://github.com/Nic30/HWToolkit',
      author='Michal Orsak',
      author_email='michal.o.socials@gmail.com',
      install_requires=[
        'simpy',  # discrete simulator
        'jinja2',  # hdl templates renderer
      ],
      license='MIT',
      packages=find_packages(),
      package_data={'hwt': ['*.vhd', '*.v']},
      include_package_data=True,
      zip_safe=False)



from simpy import Event
from simpy.core import Process
from simpy.events import URGENT, PENDING


class Initialize(Event):
    """Initializes a process. Only used internally by :class:`Process`.

    This event is automatically triggered when it is created.

    """
    def __init__(self, env, process, priority):
        # NOTE: The following initialization code is inlined from
        # Event.__init__() for performance reasons.
        self.env = env
        self.callbacks = [process._resume]
        self._value = None

        # The initialization events needs to be scheduled as urgent so that it
        # will be handled before interrupts. Otherwise a process whose
        # generator has not yet been started could be interrupted.
        self._ok = True
        env.schedule(self, priority)


class HdlProcess(Process):
    """
    Simulation process with optional start priority
    """
    def __init__(self, env, generator, priority=URGENT):
        """
        Original __init__ with new Initialize class
        """
        if not hasattr(generator, 'throw'):
            # Implementation note: Python implementations differ in the
            # generator types they provide. Cython adds its own generator type
            # in addition to the CPython type, which renders a type check
            # impractical. To workaround this issue, we check for attribute
            # name instead of type and optimistically assume that all objects
            # with a ``throw`` attribute are generators (the more intuitive
            # name ``__next__`` cannot be used because it was renamed from
            # ``next`` in Python 2).
            # Remove this workaround if it causes issues in production!
            raise ValueError('%s is not a generator.' % generator)

        # NOTE: The following initialization code is inlined from
        # Event.__init__() for performance reasons.
        self.env = env
        self.callbacks = []
        self._value = PENDING

        self._generator = generator

        # Schedule the start of the execution of the process.
        self._target = Initialize(env, self, priority)




from collections import namedtuple

from hwt.synthesizer.param import evalParam


"""
Serializer mode specifies if hdl objects derived from parent unit
should be serialized to target HDL or not

use serialize* methods to specify serialization mode for unit class

.. code-block:: python

    @serializeExclude
    class MyUnit(Unit):
        # ...
        pass

"""


def freeze_dict(data):
    keys = sorted(data.keys())
    frozen_type = namedtuple(''.join(keys), keys)
    return frozen_type(**data)


def paramsToValTuple(unit):
    d = {}
    for p in unit._params:
        name = p.getName(unit)
        v = evalParam(p)
        d[name] = v
    return freeze_dict(d)


def prepareEntity(ent, name, templateUnit):
    ent.name = name
    ent.generics.sort(key=lambda x: x.name)
    ent.ports.sort(key=lambda x: x.name)
    # copy names
    if templateUnit is not None:
        # sort in python is stable, ports and generic were added in same order
        # templateUnit should have generic and ports sorted
        for gp, gch in zip(templateUnit._entity.generics, ent.generics):
            gch.name = gp.name
        for pp, pch in zip(templateUnit._entity.ports, ent.ports):
            pch.name = pp.name


def serializeExclude(cls):
    """
    Never serialize HDL objects from this class
    """
    cls._serializeDecision = staticmethod(_serializeExclude_eval)
    return cls


def serializeOnce(cls):
    """
    Serialize HDL objects only once per class
    """
    cls._serializeDecision = staticmethod(_serializeOnce_eval)
    return cls


def serializeParamsUniq(cls):
    """
    Decide to serialize only when parameters are unique
    """
    cls._serializeDecision = staticmethod(_serializeParamsUniq_eval)
    return cls


def _serializeExclude_eval(parentUnit, obj, isDeclaration, priv):
    """
    Always decide not to serialize obj

    :param priv: private data for this function first unit of this class
    :return: tuple (do serialize this object, next priv)
    """
    if isDeclaration:
        # prepare entity which will not be serialized
        prepareEntity(obj, parentUnit.__class__.__name__, priv)

    if priv is None:
        priv = parentUnit

    return False, priv


def _serializeOnce_eval(parentUnit, obj, isDeclaration, priv):
    """
    Decide to serialize only first obj of it's class

    :param priv: private data for this function
        (first object with class == obj.__class__)

    :return: tuple (do serialize this object, next priv)
        where priv is private data for this function
        (first object with class == obj.__class__)
    """
    clsName = parentUnit.__class__.__name__

    if isDeclaration:
        obj.name = clsName

    if priv is None:
        priv = parentUnit
    elif isDeclaration:
        # prepare entity which will not be serialized
        prepareEntity(obj, clsName, priv)

    serialize = priv is obj

    return serialize, priv


def _serializeParamsUniq_eval(parentUnit, obj, isDeclaration, priv):
    """
    Decide to serialize only objs with uniq parameters and class

    :param priv: private data for this function
        ({frozen_params: obj})

    :return: tuple (do serialize this object, next priv)
    """

    params = paramsToValTuple(parentUnit)

    if priv is None:
        priv = {}

    if isDeclaration:
        try:
            prevUnit = priv[params]
        except KeyError:
            priv[params] = parentUnit
            return True, priv

        prepareEntity(obj, prevUnit._entity.name, prevUnit)
        return False, priv

    return priv[params] is parentUnit, priv



from hwt.hdl.operatorDefs import AllOps


class HwtSerializer_ops():
    opPrecedence = {
        AllOps.NOT: 4,
        AllOps.NEG: 4,
        AllOps.RISING_EDGE: 1,
        AllOps.DIV: 4,
        AllOps.ADD: 5,
        AllOps.SUB: 5,
        AllOps.MUL: 4,
        AllOps.XOR: 9,
        AllOps.EQ: 10,
        AllOps.NEQ: 10,
        AllOps.AND: 10,
        AllOps.OR: 10,
        AllOps.DOWNTO: 1,
        AllOps.GT: 10,
        AllOps.LT: 10,
        AllOps.GE: 10,
        AllOps.LE: 10,
        AllOps.CONCAT: 1,
        AllOps.INDEX: 1,
        AllOps.TERNARY: 1,
        AllOps.CALL: 1,
    }
    _binOps = {
        AllOps.AND: '%s & %s',
        AllOps.OR: '%s | %s',
        AllOps.XOR: '%s ^ %s',
        AllOps.CONCAT: "Concat(%s, %s)",
        AllOps.DIV: '%s / %s',
        AllOps.DOWNTO: "%s:%s",
        AllOps.EQ: '%s._eq(%s)',
        AllOps.GT: '%s > %s',
        AllOps.GE: '%s >= %s',
        AllOps.LE: '%s <= %s',
        AllOps.LT: '%s < %s',
        AllOps.SUB: '%s - %s',
        AllOps.MUL: '%s * %s',
        AllOps.NEQ: '%s != %s',
        AllOps.ADD: '%s + %s',
        AllOps.POW: "power(%s, %s)",
    }
    _unaryOps = {
        AllOps.NOT: "~%s",
        AllOps.NEG: "-%s",
        AllOps.RISING_EDGE: "(%s)._onRisingEdge()",
        AllOps.FALLIGN_EDGE: "(%s)._onFallingEdge()",
        AllOps.BitsAsSigned: "(%s)._convSign(True)",
        AllOps.BitsAsUnsigned: "(%s)._convSign(False)",
        AllOps.BitsAsVec: "(%s)._convSign(None)",
    }

    @classmethod
    def Operator(cls, op, ctx):
        ops = op.operands
        o = op.operator

        op_str = cls._unaryOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx))

        op_str = cls._binOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx),
                             cls._operand(ops[1], o, ctx))

        if o == AllOps.INDEX:
            assert len(ops) == 2
            return "(%s)[%s]" % (cls.asHdl(ops[0], ctx),
                                 cls._operand(ops[1], o, ctx))
        elif o == AllOps.TERNARY:
            return "(%s)._ternary(%s, %s)" %\
                tuple(map(lambda x: cls.asHdl(x, ctx), ops))
        elif o == AllOps.BitsToInt:
            assert len(ops) == 1
            op = ops[0]
            return "%s.auto_cast(INT)" % (
                cls.asHdl(op, ctx))
        elif o == AllOps.IntToBits:
            assert len(ops) == 1
            resT = op.result._dtype
            return "%s.auto_cast(%s)" % (
                cls.asHdl(ops[0], ctx),
                cls.HdlType_bits(resT, ctx))
        else:
            raise NotImplementedError(
                "Do not know how to convert %s to simModel" % (o))



from enum import Enum


class INTF_DIRECTION(Enum):
    """
    Interface direction, used in interface direction resolving process.
    """
    MASTER = 0
    SLAVE = 1
    TRISTATE = 2
    UNKNOWN = 3

    @classmethod
    def asDirection(cls, val):
        return INTF_DIRECTION_asDirecton[val]

    @classmethod
    def opposite(cls, d):
        return INTF_DIRECTION_opposite[d]


class DIRECTION(Enum):
    """
    Used to describe direction of signal.
    """
    IN = 0
    OUT = 1
    INOUT = 2

    @classmethod
    def asIntfDirection(cls, d):
        return DIRECTION_asIntfDirection[d]

    @classmethod
    def opposite(cls, d):
        return DIRECTION_opposite[d]


INTF_DIRECTION_opposite = {
    INTF_DIRECTION.SLAVE: INTF_DIRECTION.MASTER,
    INTF_DIRECTION.MASTER: INTF_DIRECTION.SLAVE,
    INTF_DIRECTION.TRISTATE: INTF_DIRECTION.TRISTATE,
}


INTF_DIRECTION_asDirecton = {
    INTF_DIRECTION.SLAVE: DIRECTION.IN,
    INTF_DIRECTION.MASTER: DIRECTION.OUT,
    INTF_DIRECTION.TRISTATE: DIRECTION.INOUT,
}


DIRECTION_asIntfDirection = {
    DIRECTION.IN: INTF_DIRECTION.SLAVE,
    DIRECTION.OUT: INTF_DIRECTION.MASTER,
    DIRECTION.INOUT: INTF_DIRECTION.TRISTATE,
}


DIRECTION_opposite = {
    DIRECTION.IN: DIRECTION.OUT,
    DIRECTION.OUT: DIRECTION.IN,
    DIRECTION.INOUT: DIRECTION.INOUT,
}


READ = "READ"
WRITE = "WRITE"
NOP = "NOP"


class Time():
    """
    Time units used mainly by simulator
    """
    ps = 1
    ns = 1000
    us = ns * 1000
    ms = us * 1000
    s = ms * 1000


class SENSITIVITY(Enum):
    """
    Sensitivity used in sensitivity resolver
    """
    ANY = 0b11
    RISING = 0b01
    FALLING = 0b10



import os
from os.path import relpath
import shutil
from typing import List

from hwt.serializer.ip_packager.component import Component
from hwt.serializer.ip_packager.helpers import prettify
from hwt.serializer.ip_packager.tclGuiBuilder import GuiBuilder,\
    paramManipulatorFns
from hwt.serializer.vhdl.serializer import VhdlSerializer
from hwt.synthesizer.uniqList import UniqList
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.utils import toRtl


# [TODO] memory maps https://forums.xilinx.com/t5/Embedded-Processor-System-Design/exporting-AXI-BASEADDR-to-xparameters-h-from-Vivado-IP/td-p/428650
class Packager(object):
    """
    Ipcore packager
    """

    def __init__(self, topUnit: Unit, name: str=None,
                 extraVhdlFiles: List[str]=[],
                 extraVerilogFiles: List[str]=[],
                 serializer=VhdlSerializer):
        assert not topUnit._wasSynthetised()
        self.topUnit = topUnit
        self.serializer = serializer
        if name:
            self.name = name
        else:
            self.name = self.topUnit._getDefaultName()

        self.hdlFiles = UniqList()

        for f in extraVhdlFiles:
            self.hdlFiles.append(f)

        for f in extraVerilogFiles:
            self.hdlFiles.append(f)

    def saveHdlFiles(self, srcDir):
        path = os.path.join(srcDir, self.name)
        try:
            os.makedirs(path)
        except OSError:
            # wipe if exists
            shutil.rmtree(path)
            os.makedirs(path)

        files = self.hdlFiles
        self.hdlFiles = toRtl(self.topUnit,
                              saveTo=path,
                              name=self.name,
                              serializer=self.serializer)

        for srcF in files:
            dst = os.path.join(path,
                               os.path.relpath(srcF, srcDir).replace('../', '')
                               )
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            shutil.copy(srcF, dst)
            self.hdlFiles.append(dst)

    def mkAutoGui(self):
        gui = GuiBuilder()
        p0 = gui.page("Main")
        handlers = []
        for g in self.topUnit._entity.generics:
            p0.param(g.name)
            for fn in paramManipulatorFns(g.name):
                handlers.append(fn)

        with open(self.guiFile, "w") as f:
            s = gui.asTcl() + '\n\n'.join(map(lambda x: str(x), handlers))
            f.write(s)

    def createPackage(self, repoDir, vendor="hwt", library="mylib",
                      description=None):
        '''
        synthetise hdl if needed
        copy hdl files
        create gui file
        create component.xml
        '''
        ip_dir = os.path.join(repoDir, self.name + "/")
        if os.path.exists(ip_dir):
            shutil.rmtree(ip_dir)

        ip_srcPath = os.path.join(ip_dir, "src")
        tclPath = os.path.join(ip_dir, "xgui")
        guiFile = os.path.join(tclPath, "gui.tcl")
        for d in [ip_dir, ip_srcPath, tclPath]:
            os.makedirs(d)
        self.saveHdlFiles(ip_srcPath)

        self.guiFile = guiFile
        self.mkAutoGui()

        c = Component()
        c._files = [relpath(p, ip_dir) for p in sorted(self.hdlFiles)] + \
                   [relpath(guiFile, ip_dir)]

        c.vendor = vendor
        c.library = library
        if description is None:
            c.description = self.name + "_v" + c.version
        else:
            c.description = description

        c.asignTopUnit(self.topUnit)

        xml_str = prettify(c.xml())
        with open(ip_dir + "component.xml", "w") as f:
            f.write(xml_str)

# def packageMultipleProjects(workspace, names, ipRepo):
#    for folder, name in names.items():
#        packageVivadoHLSProj(os.path.join(workspace, folder), "solution1", name + ".vhd", ipRepo)
#        print(folder + " packaged")
#
# def packageVivadoHLSProj(projPath, solutionName, mainVhdlFileName, ipRepo):
#    # rm others ip in project
#    vhdlPath = os.path.join(projPath, solutionName, "syn/vhdl")
#    e = entityFromFile(os.path.join(vhdlPath, mainVhdlFileName))
#    p = Packager(e, [vhdlPath])
#    p.createPackage(ipRepo)
#
# def packageBD(ipRepo, bdPath, repoPath):
#    bdName = os.path.basename(bdPath)
#    bdSourcesDir = os.path.join(bdPath, "hdl")
#    vhldFolders = []
#    ips_path = os.path.join(bdPath, "ip/")
#    vhldFolders += [os.path.join(x[0], "synth") for x in os.walk(ips_path)]  # synth subfolder of each ip
#    vhldFolders += [bdSourcesDir]
#    # ip folder
#    vhldFolders += [os.path.join(bdPath, "../../ipshared")]
#    e = entityFromFile(os.path.join(bdSourcesDir, bdName + ".vhd"))
#    p = Packager(e, vhldFolders)
#    p.createPackage(ipRepo)



from hwt.hdl.types.integerVal import IntegerVal
from hwt.synthesizer.param import evalParam


class SimModelSerializer_types():
    """
    part of SimModelSerializer responsible for type serialization
    """

    @classmethod
    def HdlType_bits(cls, typ, ctx, declaration=False):
        assert not declaration
        if typ.signed is None:
            if not (typ.forceVector or typ.bit_length() > 1):
                return 'SIM_BIT'

        w = typ.width
        if isinstance(w, int):
            pass
        else:
            w = evalParam(w)
            assert isinstance(w, IntegerVal)
            assert w._isFullVld()
            w = w.val

        return "simBitsT(%d, %r)" % (w, typ.signed)

    @classmethod
    def HdlType_bool(cls, typ, ctx, declaration=False):
        assert not declaration
        return "BOOL"

    @classmethod
    def HdlType_enum(cls, typ, ctx, declaration=False):
        if declaration:
            typ.name = ctx.scope.checkedName(typ.name, typ)

            return '%s = HEnum( "%s", [%s])' % (
                typ.name,
                typ.name,
                ", ".join(map(lambda x: '"%s"' % x,
                              typ._allValues)))
        else:
            return typ.name

    @classmethod
    def HdlType_array(cls, typ, ctx, declaration=False):
        assert not declaration
        return "HArray(%s, %d)" % (cls.HdlType(typ.elmType, ctx,
                                               declaration=declaration),
                                   evalParam(typ.size).val)



from hwt.bitmask import mask
from hwt.hdl.constants import DIRECTION
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, BIT
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.value import GenericSerializer_Value
from hwt.serializer.generic.indent import getIndent
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class VerilogSerializer_Value(GenericSerializer_Value):

    @classmethod
    def BitLiteral(cls, v, vldMask):
        if vldMask:
            return "1'b%d" % int(bool(v))
        else:
            return "1'bx"

    @classmethod
    def BitString(cls, v, width, vldMask=None):
        if vldMask is None:
            vldMask = mask(width)
        # if can be in hex
        if width % 4 == 0 and vldMask == (1 << width) - 1:
            return ("%d'h%0" + str(width // 4) + 'x') % (width, v)
        else:  # else in binary
            return cls.BitString_binary(v, width, vldMask)

    @staticmethod
    def BitString_binary(v, width, vldMask=None):
        buff = ["%d'b" % width]
        for i in range(width - 1, -1, -1):
            mask = (1 << i)
            b = v & mask

            if vldMask & mask:
                s = "1" if b else "0"
            else:
                s = "x"
            buff.append(s)
        return ''.join(buff)

    @classmethod
    def Bool_valAsHdl(cls, dtype, val, ctx):
        return str(int(val.val))

    @classmethod
    def DIRECTION(cls, d):
        if d is DIRECTION.IN:
            return "input"
        elif d is DIRECTION.OUT:
            return "output"
        elif d is DIRECTION.INOUT:
            return "inout"
        else:
            raise NotImplementedError(d)

    @classmethod
    def condAsHdl(cls, cond, forceBool, createTmpVarFn):
        if isinstance(cond, RtlSignalBase):
            cond = [cond]
        else:
            cond = list(cond)
        if len(cond) == 1:
            c = cond[0]
            if not forceBool or c._dtype == BOOL:
                return cls.asHdl(c, createTmpVarFn)
            elif c._dtype == BIT:
                return "(%s)==%s" % (cls.asHdl(c, createTmpVarFn),
                                     cls.BitLiteral(1, 1))
            elif isinstance(c._dtype, Bits):
                width = c._dtype.bit_length()
                return "(%s)!=%s" % (cls.asHdl(c, createTmpVarFn),
                                     cls.BitString(0, width))
            else:
                raise NotImplementedError()
        else:
            return " && ".join(map(lambda x: cls.condAsHdl(x, forceBool,
                                                           createTmpVarFn),
                                   cond))

    @classmethod
    def HEnumValAsHdl(cls, dtype, val, ctx):
        i = dtype._allValues.index(val.val)
        assert i >= 0
        return '%d' % i

    @classmethod
    def SignalItem(cls, si, ctx, declaration=False):
        if declaration:
            ctx = ctx.forSignal(si)

            v = si.defaultVal
            if si.virtualOnly:
                pass
            elif si.drivers:
                pass
            elif si.endpoints or si.simSensProcs:
                if not v.vldMask:
                    raise SerializerException(
                        "Signal %s is constant and has undefined value"
                        % si.name)
            else:
                raise SerializerException(
                    "Signal %s should be declared but it is not used"
                    % si.name)

            t = si._dtype
            dimensions = []
            while isinstance(t, HArray):
                # collect array dimensions
                dimensions.append(t.size)
                t = t.elmType

            s = "%s%s %s" % (getIndent(ctx.indent),
                             cls.HdlType(t, ctx),
                             si.name)
            if dimensions:
                # to make a space between name and dimensoins
                dimensions = list(map(lambda x: "[%s-1:0]"
                                                % cls.asHdl(toHVal(x), ctx),
                                      dimensions))
                dimensions.append("")
                s += " ".join(reversed(dimensions))

            if isinstance(v, RtlSignalBase):
                return s + " = %s" % cls.asHdl(v, ctx)
            elif isinstance(v, Value):
                if si.defaultVal.vldMask:
                    return s + " = %s" % cls.Value(si.defaultVal, ctx)
                else:
                    return s
            else:
                raise NotImplementedError(v)

        else:
            if si.hidden and hasattr(si, "origin"):
                return cls.asHdl(si.origin, ctx)
            else:
                return si.name

    @classmethod
    def Slice_valAsHdl(cls, dtype, val, ctx):
        upper = val.val[0]
        if isinstance(upper, Value):
            upper = upper - 1
            _format = "%s:%s"
        else:
            _format = "%s-1:%s"

        return _format % (cls.Value(upper, ctx), cls.Value(val.val[1], ctx))

    @classmethod
    def sensitivityListItem(cls, item, ctx, anyIsEventDependent):
        if isinstance(item, Operator):
            o = item.operator
            item = item.operands[0]
            if o is AllOps.RISING_EDGE:
                prefix = "posedge "
            elif o is AllOps.FALLIGN_EDGE:
                prefix = "negedge "
            else:
                raise NotImplementedError()
            return prefix + cls.asHdl(item, ctx)
        elif anyIsEventDependent:
            if item._dtype.negated:
                prefix = "negedge "
            else:
                prefix = "posedge "

            return prefix + cls.asHdl(item, ctx)

        return cls.asHdl(item, ctx)

    @classmethod
    def _BitString(cls, typeName, v, width, forceVector, vldMask):
        if vldMask != mask(width):
            if forceVector or width > 1:
                v = cls.BitString(v, width, vldMask)
            else:
                v = cls.BitLiteral(v, width, vldMask)
        else:
            v = str(v)
        # [TODO] parametrized width
        if typeName:
            return "%s(%s)" % (typeName, v)
        else:
            return v

    @classmethod
    def SignedBitString(cls, v, width, forceVector, vldMask):
        return cls._BitString("$signed", v, width, forceVector, vldMask)

    @classmethod
    def UnsignedBitString(cls, v, width, forceVector, vldMask):
        return cls._BitString("", v, width, forceVector, vldMask)



from hwt.hdl.architecture import Architecture
from hwt.hdl.entity import Entity
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bool import HBool
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.integer import Integer
from hwt.hdl.value import Value
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.exceptions import UnsupportedEventOpErr
from hwt.serializer.generic.context import SerializerCtx
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.generic.nameScope import NameScope
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class GenericSerializer():
    """
    Base class for serializers
    """
    @staticmethod
    def formater(s):
        return s

    @classmethod
    def getBaseNameScope(cls):
        """
        Get root of name space
        """
        s = NameScope(False)
        s.setLevel(1)
        s[0].update(cls._keywords_dict)
        return s

    @classmethod
    def getBaseContext(cls):
        return SerializerCtx(cls.getBaseNameScope(), 0, None, None)

    @classmethod
    def asHdl(cls, obj, ctx):
        """
        Convert object to HDL string

        :param obj: object to serialize
        :param ctx: SerializerCtx instance
        """
        if isinstance(obj, RtlSignalBase):
            return cls.SignalItem(obj, ctx)
        elif isinstance(obj, Value):
            return cls.Value(obj, ctx)
        else:
            try:
                serFn = getattr(cls, obj.__class__.__name__)
            except AttributeError:
                raise SerializerException("Not implemented for", obj)
            return serFn(obj, ctx)

    @classmethod
    def Entity_prepare(cls, ent, ctx, serialize=True):
        if serialize:
            serializedGenerics = []
            serializedPorts = []

        scope = ctx.scope
        ent.generics.sort(key=lambda x: x.name)
        ent.ports.sort(key=lambda x: x.name)

        ent.name = scope.checkedName(ent.name, ent, isGlobal=True)
        for g in ent.generics:
            g.name = scope.checkedName(g.name, g)
            if serialize:
                serializedGenerics.append(cls.GenericItem(g, ctx))

        for p in ent.ports:
            p.name = scope.checkedName(p.name, p)
            p.getSigInside().name = p.name
            if serialize:
                serializedPorts.append(cls.PortItem(p, ctx))

        if serialize:
            return serializedGenerics, serializedPorts

    @classmethod
    def Entity(cls, ent, ctx):
        """
        Entity is just forward declaration of Architecture, it is not used
        in most HDL languages as there is no recursion in hierarchy
        """

        ent.name = ctx.scope.checkedName(ent.name, ent, isGlobal=True)
        return ""

    @classmethod
    def serializationDecision(cls, obj, serializedClasses,
                              serializedConfiguredUnits):
        """
        Decide if this unit should be serialized or not eventually fix name
        to fit same already serialized unit

        :param obj: object to serialize
        :param serializedClasses: dict {unitCls : unitobj}
        :param serializedConfiguredUnits: (unitCls, paramsValues) : unitObj
            where paramsValues are named tuple name:value
        """
        isDeclaration = isinstance(obj, Entity)
        isDefinition = isinstance(obj, Architecture)
        if isDeclaration:
            unit = obj.origin
        elif isDefinition:
            unit = obj.entity.origin
        else:
            return True

        assert isinstance(unit, Unit)
        sd = unit._serializeDecision
        if sd is None:
            return True
        else:
            prevPriv = serializedClasses.get(unit.__class__, None)
            seriazlize, nextPriv = sd(unit, obj, isDeclaration, prevPriv)
            serializedClasses[unit.__class__] = nextPriv
            return seriazlize

    @classmethod
    def HdlType(cls, typ, ctx, declaration=False):
        """
        Serialize HdlType instance
        """
        if isinstance(typ, Bits):
            sFn = cls.HdlType_bits
        elif isinstance(typ, HEnum):
            sFn = cls.HdlType_enum
        elif isinstance(typ, HArray):
            sFn = cls.HdlType_array
        elif isinstance(typ, Integer):
            sFn = cls.HdlType_int
        elif isinstance(typ, HBool):
            sFn = cls.HdlType_bool
        else:
            raise NotImplementedError("type declaration is not implemented"
                                      " for type %s"
                                      % (typ.name))

        return sFn(typ, ctx, declaration=declaration)

    @classmethod
    def IfContainer(cls, ifc, ctx):
        """
        Srialize IfContainer instance
        """
        childCtx = ctx.withIndent()

        def asHdl(statements):
            return [cls.asHdl(s, childCtx) for s in statements]

        try:
            cond = cls.condAsHdl(ifc.cond, True, ctx)
        except UnsupportedEventOpErr as e:
            cond = None

        if cond is None:
            assert not ifc.elIfs
            assert not ifc.ifFalse
            stmBuff = [cls.asHdl(s, ctx) for s in ifc.ifTrue]
            return "\n".join(stmBuff)

        elIfs = []
        ifTrue = ifc.ifTrue
        ifFalse = ifc.ifFalse
        for c, statements in ifc.elIfs:
            try:
                elIfs.append((cls.condAsHdl(c, True, ctx), asHdl(statements)))
            except UnsupportedEventOpErr as e:
                if len(ifc.elIfs) == 1 and not ifFalse:
                    # register expression is in valid format and this
                    # is just register with asynchronous reset or etc...
                    ifFalse = statements
                else:
                    raise e

        return cls.ifTmpl.render(
            indent=getIndent(ctx.indent),
            cond=cond,
            ifTrue=asHdl(ifTrue),
            elIfs=elIfs,
            ifFalse=asHdl(ifFalse))

    @classmethod
    def SwitchContainer(cls, sw, ctx):
        childCtx = ctx.withIndent(1)

        def asHdl(statements):
            return [cls.asHdl(s, childCtx) for s in statements]

        switchOn = cls.condAsHdl(sw.switchOn, False, ctx)

        cases = []
        for key, statements in sw.cases:
            key = cls.asHdl(key, ctx)

            cases.append((key, asHdl(statements)))

        if sw.default:
            cases.append((None, asHdl(sw.default)))

        return cls.switchTmpl.render(
            indent=getIndent(ctx.indent),
            switchOn=switchOn,
            cases=cases)

    @classmethod
    def _operand(cls, operand, operator, ctx):
        s = cls.asHdl(operand, ctx)
        if isinstance(operand, RtlSignalBase):
            try:
                o = operand.singleDriver()
                if o.operator != operator and\
                        cls.opPrecedence[o.operator] <= cls.opPrecedence[operator]:
                    return "(%s)" % s
            except Exception:
                pass
        return s

    @classmethod
    def _bin_op(cls, operator, op_str, ctx, ops):
        return op_str.join(map(lambda operand: cls._operand(operand, operator, ctx), ops))


from hwt.hdl.constants import INTF_DIRECTION
from hwt.serializer.ip_packager.helpers import appendSpiElem, \
    mkSpiElm


class BusInterface():
    def __init__(self):
        self.name = None
        self.busType = None
        self.abstractionType = None
        self.isMaster = None
        # logical : physical
        self._portMaps = {}
        self.parameters = []
        self.endianness = None

    # @classmethod
    # def fromElem(cls, elm):
    #    self = cls()
    #    self.name = elm.find('spirit:name', ns).text
    #    self.busType = Type.fromElem(elm.find('spirit:busType', ns))
    #    self.abstractionType = Type.fromElem(elm.find('spirit:abstractionType', ns))
    #    if elm.find('spirit:master', ns) is not None:
    #        self.isMaster = True
    #    elif elm.find('spirit:slave', ns) is not None:
    #        self.isMaster = False
    #    else:
    #        raise Exception("buss missing master/slave specification")
    #    self._portMaps = []
    #    for m in elm.find('spirit:_portMaps', ns):
    #        pm = PortMap.fromElem(m)
    #        self._portMaps.append(pm)
    #
    #    self.parameters = []
    #    for p in elm.find('spirit:parameters', ns):
    #        p_obj = Parameter.fromElem(p)
    #        self.parameters.append(p_obj)
    #
    #    return self

    @staticmethod
    def generatePortMap(biType, intf):
        def processIntf(mapDict, intf):
            if not intf._interfaces:
                assert(isinstance(mapDict, str))
                return {mapDict: intf._getPhysicalName()}
            else:
                d = {}
                for i in intf._interfaces:
                    if i._isExtern:
                        try:
                            m = mapDict[i._name]
                        except KeyError:
                            raise Exception("Interface %s has interface %s which is not defined in ipcore interface class"
                                            % (intf._getFullName(), i._name)
                                            )

                        d.update(processIntf(m, i))
                return d
        return processIntf(biType.map, intf)

    @classmethod
    def fromBiClass(cls, intf, biClass):
        self = BusInterface()
        biType = biClass()
        self.name = intf._name
        self.busType = biType
        self.abstractionType = biClass()
        self.abstractionType.name += "_rtl"
        self.isMaster = intf._direction == INTF_DIRECTION.MASTER
        self._portMaps = BusInterface.generatePortMap(biType, intf)
        self.parameters = biType.parameters
        return self

    def asElem(self):
        def mkPortMap(logicalName, physicalName):
            pm = mkSpiElm("portMap")
            appendSpiElem(appendSpiElem(pm, "logicalPort"),
                          "name").text = logicalName
            appendSpiElem(appendSpiElem(pm, "physicalPort"),
                          "name").text = physicalName
            return pm

        e = mkSpiElm("busInterface")

        appendSpiElem(e, 'name').text = self.name
        e.append(self.busType.asElem('busType'))
        e.append(self.abstractionType.asElem('abstractionType'))
        if self.isMaster:
            appendSpiElem(e, "master")
        else:
            appendSpiElem(e, "slave")

        pm = appendSpiElem(e, "portMaps")

        for lName, pName in sorted(self._portMaps.items(),
                                   key=lambda pm: pm[0]):
            pm.append(mkPortMap(lName, pName))
        if self.endianness is not None:
            appendSpiElem(e, "endianness").text = self.endianness
        if len(self.parameters) > 0:
            pm = appendSpiElem(e, "parameters")
            for p in self.parameters:
                pElm = p.asElem()
                pm.append(pElm)
        return e



from typing import Tuple

from hwt.hdl.transPart import TransPart
from hwt.hdl.transTmpl import OneOfTransaction


def iterSort(iterators, cmpFn):
    """
    Sort items from iterators(generators) by alwas selecting item
    with lowest value (min first)

    :return: generator of tuples (origin index, item) where origin index
        is index of iterator in "iterators" from where item commes from
    """
    actual = []
    _iterators = []
    for i, it in enumerate(iterators):
        try:
            a = next(it)
            _iterators.append((i, it))
            actual.append(a)
        except StopIteration:
            continue

    while True:
        if not _iterators:
            return
        elif len(_iterators) == 1:
            originIndex, it = _iterators[0]
            yield originIndex, actual[0]
            for item in it:
                yield originIndex, item
            return

        # select minimum and iterator from where it comes from
        minimum = None
        minimumIndex = None
        secondMin = None
        for i, val in enumerate(actual):
            skipSecMinCheck = False
            if minimum is None:
                minimum = val
                minimumIndex = i
            elif cmpFn(val, minimum):
                secondMin = minimum
                minimum = val
                minimumIndex = i
                skipSecMinCheck = True
            elif not skipSecMinCheck and (
                    secondMin is None or cmpFn(val, secondMin)):
                secondMin = val

        actualI, actualIt = _iterators[minimumIndex]
        while not cmpFn(secondMin, minimum):
            yield (actualI, minimum)
            try:
                minimum = next(actualIt)
            except StopIteration:
                minimum = None
                break

        # consume from actual iterator while
        if minimum is None:
            del _iterators[minimumIndex]
            del actual[minimumIndex]
        else:
            # minimum is not minimum anymore
            actual[minimumIndex] = minimum


class ChoicesOfFrameParts(list):
    """
    List of ChoiceOfFrameParts
    One of ChoiceOfFrameParts is used to represent the word, item depends
    on context

    :ivar origin: OneOfTransaction instance
    :ivar startOfPart: bit addr of start of this group of frame parts
    :ivar endOfPart: bit addr of end of this group of frame parts
    :ivar _isLast: flag which means this is the last part of original union
    """

    def __init__(self, startOfPart: int, origin: OneOfTransaction):
        self.origin = origin
        self.startOfPart = startOfPart
        self.endOfPart = None
        self._isLast = False

    def resolveEnd(self):
        end = self.startOfPart
        for items in self:
            if items:
                end = max(end, max(itm.endOfPart for itm in items))
        self.endOfPart = end

    def setIsLast(self, val: bool) -> None:
        self._isLast = val

    def isLastPart(self) -> bool:
        """
        :return: True if this part is last in parts derived from original field
            else False
        """
        return self._isLast

    def getBusWordBitRange(self) -> Tuple[int, int]:
        """
        :return: bit range which contains data of this part on bus data signal
        """
        offset = self.startOfPart % self.parent.wordWidth
        return (offset + self.bit_length(), offset)

    def bit_length(self):
        return self.endOfPart - self.startOfPart

    def __repr__(self):
        return "<ChoicesOfFrameParts %s>" % list.__repr__(self)


class ChoiceOfFrameParts(list):
    """
    :ivar origin: ChoicesOfFrameParts instance
    :ivar tmpl: TransTmpl which was item generated from
    """

    def __init__(self, origin, tmpl):
        self.origin = origin
        self.tmpl = tmpl

    def __repr__(self):
        return "<ChoiceOfFrameParts %s>" % list.__repr__(self)


def groupIntoChoices(splitsOnWord, wordWidth: int, origin: OneOfTransaction):
    """
    :param splitsOnWord: list of lists of parts (fields splited on word
        boundaries)
    :return: generators of ChoicesOfFrameParts for each word
        which are not crossing word boundaries
    """
    def cmpWordIndex(a, b):
        return a.startOfPart // wordWidth < b.startOfPart // wordWidth

    actual = None
    itCnt = len(splitsOnWord)
    for i, item in iterSort(splitsOnWord, cmpWordIndex):
        _actualW = item.startOfPart // wordWidth
        if actual is None:
            actual = ChoicesOfFrameParts(item.startOfPart, origin)
            actual.extend(
                ChoiceOfFrameParts(actual,
                                   origin.possibleTransactions[_i])
                for _i in range(itCnt))
            actualW = _actualW
        elif _actualW > actualW:
            actual.resolveEnd()
            yield actual
            actual = ChoicesOfFrameParts(item.startOfPart, origin)
            actual.extend(
                ChoiceOfFrameParts(actual,
                                   origin.possibleTransactions[_i])
                for _i in range(itCnt))
            actualW = _actualW
        actual[i].append(item)

    if actual is not None:
        actual.setIsLast(True)
        actual.resolveEnd()
        yield actual


class TransTmplWordIterator():
    """
    Iterator which reinterprets any structure
    as generator of bits of specified width
    """
    def __init__(self, wordWidth):
        self.wordWidth = wordWidth

    def fullWordCnt(self, start: int, end: int):
        """Count of complete words between two addresses
        """
        assert end >= start, (start, end)
        gap = max(0, (end - start) - (start % self.wordWidth))
        return gap // self.wordWidth

    def groupByWordIndex(self, transaction: 'TransTmpl', offset: int):
        """
        Group transaction parts splited on words to words

        :param transaction: TransTmpl instance which parts
            should be grupped into words
        :return: generator of tuples (wordIndex, list of transaction parts
            in this word)
        """
        actualW = None
        partsInWord = []
        wordWidth = self.wordWidth
        for item in self.splitOnWords(transaction, offset):
            _actualW = item.startOfPart // wordWidth
            if actualW is None:
                actualW = _actualW
                partsInWord.append(item)
            elif _actualW > actualW:
                yield (actualW, partsInWord)
                actualW = _actualW
                partsInWord = [item, ]
            else:
                partsInWord.append(item)

        if partsInWord:
            yield (actualW, partsInWord)

    def splitOnWords(self, transaction, addrOffset=0):
        """
        :return: generator of TransPart instance
        """
        wordWidth = self.wordWidth
        end = addrOffset
        for tmp in transaction.walkFlatten(offset=addrOffset):
            if isinstance(tmp, OneOfTransaction):
                split = [self.splitOnWords(ch, end)
                         for ch in tmp.possibleTransactions]
                yield from groupIntoChoices(split, wordWidth, tmp)
                end = addrOffset + tmp.possibleTransactions[0].bitAddrEnd
            else:
                (base, end), tmpl = tmp
                startOfPart = base
                while startOfPart != end:
                    wordIndex = startOfPart // wordWidth
                    endOfWord = (wordIndex + 1) * wordWidth
                    endOfPart = min(endOfWord, end)
                    inFieldOffset = startOfPart - base
                    yield TransPart(self, tmpl, startOfPart, endOfPart,
                                    inFieldOffset)
                    startOfPart = endOfPart



from hwt.hdl.assignment import Assignment
from hwt.hdl.variables import SignalItem
from hwt.serializer.generic.indent import getIndent


class DebugTmpVarStack():
    """
    Collect tmp variables in serialization and format them to readable form

    .. code-block:: python
        tmpVars = DebugTmpVarStack()
        ctx = VhdlSerializer.getBaseContext()
        ctx.createTmpVarFn = tmpVars.createTmpVarFn
        s = VhdlSerializer.Assignment(self, ctx)
        "%s%s" % (tmpVars.serialize(), s)

    """
    def __init__(self, serializer):
        """
        :ivar vars: list of serialized variable declarations
        """
        self.vars = []
        self.serializer = serializer
        self.ctx = self.serializer.getBaseContext()

    def createTmpVarFn(self, suggestedName, dtype):
        # [TODO] it is better to use RtlSignal
        ser = self.serializer
        s = SignalItem(suggestedName, dtype, virtualOnly=True)
        s.hidden = False
        serializedS = ser.SignalItem(s, self.ctx, declaration=True)
        self.vars.append((serializedS, s))

        return s

    def _serializeItem(self, item):
        var, s = item
        # assignemt of value for this tmp variable
        a = Assignment(s.defaultVal, s, virtualOnly=True)
        return "%s\n%s" % (var, self.serializer.Assignment(a, self.ctx))

    def serialize(self, indent=0):
        if not self.vars:
            return ""

        separator = getIndent(indent) + "\n"
        return separator.join(map(self._serializeItem, self.vars)) + "\n"



from typing import Callable, Tuple, Generator, Union, Optional

from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.struct import HStruct, HStructField
from hwt.hdl.types.union import HUnion
from hwt.pyUtils.arrayQuery import iter_with_last
from hwt.synthesizer.param import evalParam


def _default_shouldEnterFn(transTmpl: 'TransTmpl') -> Tuple[bool, bool]:
    return (bool(transTmpl.children), not bool(transTmpl.children))


class TransTmpl(object):
    """
    Container of informations about frames generated from any HType
    (HStruct etc.)

    * contains precalculated address range for all members of type

    :ivar dtype: type of this item
    :ivar bitAddr: offset of start of this item in bits
    :ivar parent: object which generated this item, optional TransTmpl
    :ivar origin: object which was template for generating of this item
    :ivar itemCnt: if this transaction template is for arry this is
        item count for array
    :ivar childrenAreChoice: flag which tells if childrens are sequence
        or only one of them can be used in same time
    """

    def __init__(self, dtype: HdlType, bitAddr: int=0,
                 parent: Optional['TransTmpl']=None,
                 origin: Optional[HStructField]=None):
        self.parent = parent
        if origin is None:
            origin = dtype

        self.origin = origin
        self.dtype = dtype
        self.children = []
        self._loadFromHType(dtype, bitAddr)

    def _loadFromArray(self, dtype: HdlType, bitAddr: int) -> int:
        """
        Parse HArray type to this transaction template instance
        :return: address of it's end
        """
        self.itemCnt = evalParam(dtype.size).val
        self.children = TransTmpl(
            dtype.elmType, 0, parent=self, origin=self.origin)
        return bitAddr + self.itemCnt * self.children.bitAddrEnd

    def _loadFromBits(self, dtype: HdlType, bitAddr: int):
        """
        Parse Bits type to this transaction template instance
        :return: address of it's end
        """
        return bitAddr + dtype.bit_length()

    def _loadFromHStruct(self, dtype: HdlType, bitAddr: int):
        """
        Parse HStruct type to this transaction template instance
        :return: address of it's end
        """
        for f in dtype.fields:
            t = f.dtype
            origin = f
            isPadding = f.name is None

            if isPadding:
                width = t.bit_length()
                bitAddr += width
            else:
                fi = TransTmpl(t, bitAddr, parent=self, origin=origin)
                self.children.append(fi)
                bitAddr = fi.bitAddrEnd

        return bitAddr

    def _loadFromUnion(self, dtype: HdlType, bitAddr: int) -> int:
        """
        Parse HUnion type to this transaction template instance
        :return: address of it's end
        """
        for field in dtype.fields.values():
            ch = TransTmpl(field.dtype, 0, parent=self, origin=field)
            self.children.append(ch)
        return bitAddr + dtype.bit_length()

    def _loadFromHType(self, dtype: HdlType, bitAddr: int) -> None:
        """
        Parse any HDL type to this transaction template instance
        """
        self.bitAddr = bitAddr

        if isinstance(dtype, HStruct):
            ld = self._loadFromHStruct
        elif isinstance(dtype, HArray):
            ld = self._loadFromArray
        elif isinstance(dtype, HUnion):
            ld = self._loadFromUnion
        else:
            ld = self._loadFromBits

        self.bitAddrEnd = ld(dtype, bitAddr)
        self.childrenAreChoice = isinstance(dtype, HUnion)

    def getItemWidth(self) -> int:
        """
        Only for transactions derived from HArray

        :return: width of item in original array
        """
        if not isinstance(self.dtype, HArray):
            raise TypeError()
        return (self.bitAddrEnd - self.bitAddr) // self.itemCnt

    def bit_length(self) -> int:
        """
        :return: number of bits in this transaction
        """
        return self.bitAddrEnd - self.bitAddr

    def walkFlatten(self, offset: int=0,
                    shouldEnterFn=_default_shouldEnterFn) -> Generator[
            Union[Tuple[Tuple[int, int], 'TransTmpl'], 'OneOfTransaction'],
            None, None]:
        """
        Walk fields in instance of TransTmpl

        :param offset: optional offset for all children in this TransTmpl
        :param shouldEnterFn: function (transTmpl) which returns True
            when field should be split on it's children
        :param shouldEnterFn: function(transTmpl) which should return
            (shouldEnter, shouldUse) where shouldEnter is flag that means
            iterator should look inside of this actual object
            and shouldUse flag means that this field should be used
            (=generator should yield it)
        :return: generator of tuples ((startBitAddress, endBitAddress),
            TransTmpl instance)
        """

        t = self.dtype
        base = self.bitAddr + offset
        end = self.bitAddrEnd + offset

        shouldEnter, shouldYield = shouldEnterFn(self)
        if shouldYield:
            yield ((base, end), self)

        if shouldEnter:
            if isinstance(t, Bits):
                pass
            elif isinstance(t, HStruct):
                for ch in self.children:
                    yield from ch.walkFlatten(offset=offset,
                                              shouldEnterFn=shouldEnterFn)
            elif isinstance(t, HArray):
                itemSize = (self.bitAddrEnd - self.bitAddr) // self.itemCnt
                for i in range(self.itemCnt):
                    yield from self.children.walkFlatten(
                        offset=base + i * itemSize,
                        shouldEnterFn=shouldEnterFn)
            elif isinstance(t, HUnion):
                yield OneOfTransaction(self, offset, shouldEnterFn,
                                       self.children)
            else:
                raise TypeError(t)

    def __repr__(self, offset: int=0):
        offsetStr = "".join(["    " for _ in range(offset)])

        try:
            name = self.origin.name
        except AttributeError:
            name = None

        if name:
            name = " name:%s," % name
        else:
            name = ""

        s = "%s<TransTmpl%s start:%d, end:%d" % (offsetStr, name,
                                                 self.bitAddr, self.bitAddrEnd)
        if isinstance(self.dtype, HArray):
            s += ", itemCnt:%d" % (self.itemCnt) + "\n"
            s += self.children.__repr__(offset=offset + 1) + "\n"
            s += offsetStr + ">"
            return s
        elif not self.children:
            return s + ">"

        buff = [s, ]
        for isLast, child in iter_with_last(self.children):
            buff.append(child.__repr__(offset=offset + 1))
            if self.childrenAreChoice and not isLast:
                buff.append(offsetStr + "    <OR>")

        buff.append(offsetStr + ">")
        return "\n".join(buff)


class OneOfTransaction(object):
    """
    Container of possible transactions for transactions deriverd
    from HUnion type

    :ivar parent: parent TransTmpl instance
    :ivar offset: bit addr offset in parent type structure
    :ivar shouldEnterFn: function(transTmpl) which should
        return (shouldEnter, shouldUse)
        where shouldEnter is flag that means iterator should look
        inside of this actual object and shouldUse flag means that this field
        should be used (=generator should yield it)
    :ivar possibleTransactions: tuple of TransTmpl instances
        from which only one can be used in same time
    """

    def __init__(self, parent: TransTmpl,
                 offset: int,
                 shouldEnterFn: Callable[[TransTmpl], Tuple[bool, bool]],
                 possibleTransactions: Tuple[TransTmpl]):
        self.parent = parent
        self.offset = offset
        self.shouldEnterFn = shouldEnterFn
        self.possibleTransactions = possibleTransactions

    def walkFlattenChilds(self) -> Generator[
            Union[Tuple[Tuple[int, int], TransTmpl], 'OneOfTransaction'],
            None, None]:
        """
        :return: generator of generators of tuples
            ((startBitAddress, endBitAddress), TransTmpl instance)
            for each possiblility in this transaction
        """
        for p in self.possibleTransactions:
            yield p.walkFlatten(offset=self.offset,
                                shouldEnterFn=self.shouldEnterFn)



from hwt.hdl.types.hdlType import HdlType


class HBool(HdlType):

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return hash(HBool)

    @classmethod
    def get_auto_cast_fn(cls):
        from hwt.hdl.types.boolCast import cast_hbool
        return cast_hbool

    def bit_length(self):
        return 1

    def all_mask(self):
        return 1

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.boolVal import HBoolVal
            cls._valCls = HBoolVal
            return cls._valCls



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil

from hwt.hdl.architecture import Architecture
from hwt.hdl.entity import Entity
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.serializer import GenericSerializer
from hwt.serializer.vhdl.serializer import VhdlSerializer
from hwt.synthesizer.uniqList import UniqList
from hwt.synthesizer.unit import Unit


def toRtl(unitOrCls: Unit, name: str=None,
          serializer: GenericSerializer=VhdlSerializer,
          targetPlatform=None, saveTo: str=None):
    """
    Convert unit to RTL using specified serializer

    :param unitOrCls: unit instance or class, which should be converted
    :param name: name override of top unit (if is None name is derived
        form class name)
    :param serializer: serializer which should be used for to RTL conversion
    :param targetPlatform: metainformatins about target platform, distributed
        on every unit under _targetPlatform attribute
        before Unit._impl() is called
    :param saveTo: directory where files should be stored
        If None RTL is returned as string.
    :raturn: if saveTo returns RTL string else returns list of file names
        which were created
    """
    if not isinstance(unitOrCls, Unit):
        u = unitOrCls()
    else:
        u = unitOrCls

    u._loadDeclarations()
    if name is not None:
        u._name = name

    globScope = serializer.getBaseNameScope()
    mouduleScopes = {}

    # unitCls : unitobj
    serializedClasses = {}

    # (unitCls, paramsValues) : unitObj
    # where paramsValues are dict name:value
    serializedConfiguredUnits = {}

    doSerialize = True

    createFiles = saveTo is not None
    if createFiles:
        os.makedirs(saveTo, exist_ok=True)
        files = UniqList()
    else:
        codeBuff = []

    for obj in u._toRtl(targetPlatform):
        doSerialize = serializer.serializationDecision(
            obj,
            serializedClasses,
            serializedConfiguredUnits)
        if doSerialize:
            if isinstance(obj, Entity):
                s = globScope.fork(1)
                s.setLevel(2)
                ctx = serializer.getBaseContext()
                ctx.scope = s
                mouduleScopes[obj] = ctx

                sc = serializer.Entity(obj, ctx)
                if createFiles:
                    fName = obj.name + serializer.fileExtension
                    fileMode = 'w'

            elif isinstance(obj, Architecture):
                try:
                    ctx = mouduleScopes[obj.entity]
                except KeyError:
                    raise SerializerException(
                        "Entity should be serialized"
                        " before architecture of %s"
                        % (obj.getEntityName()))

                sc = serializer.Architecture(obj, ctx)
                if createFiles:
                    fName = obj.getEntityName() + serializer.fileExtension
                    fileMode = 'a'
            else:
                if hasattr(obj, "_hdlSources"):
                    for fn in obj._hdlSources:
                        if isinstance(fn, str):
                            shutil.copy2(fn, saveTo)
                            files.append(fn)
                            continue
                else:
                    sc = serializer.asHdl(obj)

            if sc:
                if createFiles:
                    fp = os.path.join(saveTo, fName)
                    files.append(fp)
                    with open(fp, fileMode) as f:
                        if fileMode == 'a':
                            f.write("\n")

                        f.write(
                            serializer.formater(sc)
                        )
                else:
                    codeBuff.append(sc)

        elif not createFiles:
            try:
                name = '"%s"' % obj.name
            except AttributeError:
                name = ""
            codeBuff.append(serializer.comment(
                "Object of class %s, %s was not serialized as specified" % (
                    obj.__class__.__name__, name)))

    if createFiles:
        return files
    else:
        return serializer.formater(
            "\n".join(codeBuff)
        )


def serializeAsIpcore(unit, folderName=".", name=None,
                      serializer=VhdlSerializer):
    from hwt.serializer.ip_packager.packager import Packager
    p = Packager(unit, name=name)
    p.createPackage(folderName)
    return p





class Value():
    """
    Wrap around hdl value with overloaded operators
    http://www.rafekettler.com/magicmethods.html

    operators are overloaded in every type separately
    """
    __slots__ = ["val", "_dtype", "vldMask", "updateTime"]

    def __init__(self, val, dtype, vldMask, updateTime=-1):
        """
        :param val: pythonic value representing this value
        :param dtype: data type object from which this value was derived from
        :param vldMask: validity mask for value
        :param updateTime: simulation time when this value vas created
        """

        self.val = val
        self._dtype = dtype
        self.vldMask = vldMask
        self.updateTime = updateTime

    def _isFullVld(self):
        return self.vldMask == self._dtype.all_mask()

    def _auto_cast(self, toT):
        "type cast to compatible type"
        return self._dtype.auto_cast(self, toT)

    def _reinterpret_cast(self, toT):
        "type cast to type of same size"
        return self._dtype.reinterpret_cast(self, toT)

    def staticEval(self):
        return self.clone()

    def clone(self):
        return self.__class__(self.val, self._dtype, self.vldMask,
                              self.updateTime)

    def __hash__(self):
        return hash((self._dtype, self.val, self.vldMask, self.updateTime))

    def __repr__(self):
        return "<{0:s} {1:s}, mask {2:x}, time {3:.2f}>".format(
            self.__class__.__name__, str(self.val), self.vldMask,
            self.updateTime)

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        raise NotImplementedError(
            "fromPy fn is not implemented for %r" % (cls))

    def __int__(self):
        if isinstance(self, Value) or self._const:
            if self._isFullVld():
                return self.val
            else:
                raise ValueError("Value of %r is not fully defined" % self)

        raise ValueError(
            "Value of %r is not constant it can be statically solved" % self)

    def __bool__(self):
        if isinstance(self, Value) or self._const:
            if self._isFullVld():
                return bool(self.val)
            else:
                raise ValueError("Value of %r is not fully defined" % self)

        raise ValueError(
            "Value of %r is not constant it can be statically solved" % self)

    def __eq__(self, other):
        if areValues(self, other):
            return self._dtype == other._dtype and \
                self._isFullVld() and other._isFullVld() and\
                self.val == other.val
        else:
            return super().__eq__(other)

    def _eq(self, other):
        raise TypeError()

    def __ne__(self, other):
        eq = self._eq(other)
        eq.val = not eq.val
        return eq

    # def __pos__(self):
    #     raise TypeError()
    #
    # def __neg__(self):
    #     raise TypeError()
    #
    # def __abs__(self):
    #     raise TypeError()
    #
    # def __invert__(self):
    #     raise TypeError()
    #
    # def __round__(self, n):
    #     raise TypeError()
    #
    # def __floor__(self):
    #     raise TypeError()
    #
    # def __ceil__(self):
    #     raise TypeError()
    #
    # def __add__(self, other):
    #     raise TypeError()
    #
    # def __sub__(self, other):
    #     raise TypeError()
    #
    # def __mul__(self, other):
    #     raise TypeError()
    #
    # def __floordiv__(self, other):
    #     raise TypeError()
    #
    # def __div__(self, other):
    #     raise TypeError()
    #
    # def __truediv__(self, other):
    #     raise TypeError()
    #
    # def __mod__(self, other):
    #     raise TypeError()
    #
    # def __divmod__(self, other):
    #     raise TypeError()
    #
    # def __pow__(self, other):
    #     raise TypeError()
    #
    # def __lshift__(self, other):
    #     raise TypeError()
    #
    # def __rshift__(self, other):
    #     raise TypeError()
    #
    # def __and__(self, other):
    #     raise TypeError()
    #
    # def __or__(self, other):
    #     raise TypeError()
    #
    # def __xor__(self, other):
    #     raise TypeError()
    #
    # def _concat(self, other):
    #     raise TypeError()
    #
    # def __lt__(self, other):
    #     raise TypeError()
    #
    # def __le__(self, other):
    #     raise TypeError()
    #
    # def __gt__(self, other):
    #     raise TypeError()
    #
    # def __ge__(self, other):
    #     raise TypeError()
    #
    # def _onRisingEdge(self, now):
    #     raise TypeError()
    #
    # def _onFallingEdge(self, now):
    #     raise TypeError()


def areValues(*items):
    res = True
    for i in items:
        res = res and isinstance(i, Value)
    return res



import math
import os
from time import gmtime, strftime

from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, STR
from hwt.hdl.types.integer import Integer
from hwt.serializer.ip_packager.helpers import spi_ns_prefix, mkSpiElm, \
    appendSpiElem, appendStrElements, mkXiElm, appendXiElem, appendSpiAtribs
from hwt.synthesizer.param import evalParam


class Value():
    __slots__ = ['id', 'format', 'bitStringLength',
                 'resolve', 'dependency', 'text']
    RESOLVE_GENERATED = "generated"
    RESOLVE_USER = "user"

    # @classmethod
    # def fromElem(cls, elm):
    #     self = cls()
    #     self.id = elm.attrib[spi_ns_prefix + 'id']
    #     self.text = elm.text
    #     self.resolve = elm.attrib[spi_ns_prefix + 'resolve']
    #     self.dependency = elm.attrib[spi_ns_prefix + 'dependency']
    #     for n in ['format', 'bitStringLength']:
    #         try:
    #             value = elm.attrib[spi_ns_prefix + n]
    #             setattr(self, n, value)
    #         except KeyError:
    #             pass
    #     return self

    @classmethod
    def fromGeneric(cls, idPrefix, g, resolve):
        self = cls()
        self.id = idPrefix + g.name
        self.resolve = resolve
        t = g._dtype

        def getVal():
            v = evalParam(g.defaultVal)
            if v.vldMask:
                return v.val
            else:
                return 0

        def bitString(w):
            self.format = "bitString"
            digits = math.ceil(w / 4)
            self.text = ('0x%0' + str(digits) + 'X') % getVal()
            self.bitStringLength = str(w)

        if t == BOOL:
            self.format = "bool"
            self.text = str(bool(getVal())).lower()
        elif isinstance(t, Integer):
            self.format = "long"
            self.text = str(getVal())
        elif t == STR:
            self.format = "string"
            self.text = g.defaultVal.staticEval().val
        elif isinstance(t, Bits):
            bitString(g.defaultVal._dtype.bit_length())
        else:
            raise NotImplementedError(
                "Not implemented for datatype %s" % repr(t))
        return self

    def asElem(self):
        e = mkSpiElm("value")
        appendSpiAtribs(self, e, spi_ns_prefix, reqPropNames=['id', 'resolve'],
                        optPropNames=['format', 'bitStringLength'])

        e.text = str(self.text)
        return e


class FileSet():
    def __init__(self):
        self.name = ""
        self.files = []

    def asElem(self):
        e = mkSpiElm("fileSet")
        appendSpiElem(e, "name").text = self.name
        for f in self.files:
            e.append(f.asElem())
        return e


class File():
    _strValues = ["name", "fileType", "userFileType"]

    def __init__(self):
        self.name = ""
        self.fileType = ""
        self.userFileType = ""

    @classmethod
    def fromFileName(cls, fileName):
        self = cls()
        extDict = {
            ".vhd": ("vhdlSource", "IMPORTED_FILE"),
            ".tcl": ("tclSource", "XGUI_VERSION_2"),
            ".v":   ("verilogSource", "IMPORTED_FILE")
        }
        fileType = extDict[os.path.splitext(fileName.lower())[1]]
        self.fileType = fileType[0]
        self.userFileType = fileType[1]
        self.name = fileName
        return self

    def asElem(self):
        e = mkSpiElm("file")
        appendStrElements(e, self, self._strValues)
        return e


class Parameter():
    __slots__ = ["name", 'displayName', "value", 'order']

    def __init__(self):
        self.name = ""
        self.value = Value()

    # @classmethod
    # def fromElem(cls, elm):
    #    self = cls()
    #    self.name = elm.find('spirit:name', ns).text
    #    v = elm.find('spirit:value', ns)
    #    self.value = Value.fromElem(v)
    #    return self

    def asElem(self):
        e = mkSpiElm("parameter")
        appendSpiElem(e, "name").text = self.name
        e.append(self.value.asElem())
        return e


class CoreExtensions():
    def __init__(self):
        self.supportedFamilies = {
            "zynq": "Production",
            "atrix7": "Production",
            "kintex7": "Production",
            "virtex7": "Production"
        }
        self.taxonomies = [
            "/BaseIP"
        ]
        self.displayName = ""
        self.coreRevision = ""
        self.coreCreationDateTime = gmtime()

    def asElem(self, displayName, revision):
        r = mkXiElm("coreExtensions")
        sf = appendXiElem(r, "supportedFamilies")
        for family, lifeCycle in self.supportedFamilies.items():
            f = appendXiElem(sf, "family")
            f.text = family
            f.attrib["xilinx:lifeCycle"] = lifeCycle

        ta = appendXiElem(r, "taxonomies")
        for t in self.taxonomies:
            appendXiElem(ta, "taxonomy").text = t

        appendXiElem(r, "displayName").text = displayName
        appendXiElem(r, "coreRevision").text = revision
        appendXiElem(r, "coreCreationDateTime").text = strftime(
            "%Y-%m-%dT%H:%M:%SZ", self.coreCreationDateTime)
        return r


#  [TODO] XILINX_VERSION has to be extracted into some configuration
class VendorExtensions():
    XILINX_VERSION = "2014.4.1"

    def __init__(self):
        self.coreExtensions = CoreExtensions()
        self.packagingInfo = {"xilinxVersion": self.XILINX_VERSION}

    def asElem(self, displayName, revision):
        r = mkSpiElm("vendorExtensions")
        r.append(self.coreExtensions.asElem(displayName, revision))
        pi = appendXiElem(r, "packagingInfo")
        for key, val in self.packagingInfo.items():
            appendXiElem(pi, key).text = val

        return r



from _functools import reduce
from typing import List

from hwt.hdl.assignment import Assignment
from hwt.hdl.hdlObject import HdlObject
from hwt.hdl.value import Value


def seqEvalCond(cond):
    _cond = True
    for c in cond:
        _cond = _cond and bool(c.staticEval().val)

    return _cond


def isSameHVal(a, b):
    return a is b or (isinstance(a, Value)
                      and isinstance(b, Value)
                      and a.val == b.val
                      and a.vldMask == b.vldMask)


def isSameStatement(stmA, stmB):
    if isinstance(stmA, Assignment):
        if isinstance(stmB, Assignment):
            if isSameHVal(stmA.dst, stmB.dst)\
                    and isSameHVal(stmA.src, stmB.src):
                return True
        return False
    else:
        return stmA.isSame(stmB)


def isSameStatementList(stmListA, stmListB):
    for a, b in zip(stmListA, stmListB):
        if not isSameStatement(a, b):
            return False
    return True


def statementsAreSame(statements):
    iterator = iter(statements)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(isSameStatement(first, rest) for rest in iterator)


class IfContainer(HdlObject):
    """
    Structural container of if statement for hdl rendering
    """

    def __init__(self, cond, ifTrue=[], ifFalse=[], elIfs=[]):
        """
        :param cond: list of conditions for this if
        :param ifTrue: list of statements which should be active if cond.
            is met
        :param elIfs: list of tuples (list of conditions, list of statements)
        :param ifFalse: list of statements which should be active if cond.
            and any other cond. in elIfs is met
        """
        self.cond = cond
        self.ifTrue = ifTrue
        self.elIfs = elIfs
        self.ifFalse = ifFalse

    @classmethod
    def potentialyReduced(cls, cond, ifTrue=[], ifFalse=[], elIfs=[])\
            -> List[HdlObject]:
        """
        If conditions have no effect on result
        IfContainer is reduced to just list of assignments

        Params same as `IfContainer.__init__`
        """
        if IfContainer.condHasEffect(ifTrue, ifFalse, elIfs):
            return [IfContainer(cond, ifTrue, ifFalse, elIfs), ]
        else:
            return ifTrue

    @classmethod
    def condHasEffect(cls, ifTrue, ifFalse, elIfs):
        stmCnt = len(ifTrue)
        if stmCnt == len(ifFalse) and reduce(lambda x, y: x and y,
                                             [len(stm) == stmCnt
                                              for _, stm in elIfs],
                                             True):
            for stms in zip(ifTrue, ifFalse, *map(lambda x: x[1], elIfs)):
                if not statementsAreSame(stms):
                    return True
            return False
        return True

    def isSame(self, other):
        """
        :return: True if other has same meaning as this statement
        """
        if isinstance(other, IfContainer):
            if self.cond == other.cond:
                if len(self.ifTrue) == len(other.ifTrue) \
                        and len(self.ifFalse) == len(other.ifFalse) \
                        and len(self.elIfs) == len(other.elIfs):
                    if not isSameStatementList(self.ifTrue,
                                               other.ifTrue) \
                            or not isSameStatementList(self.ifFalse,
                                                       other.ifFalse):
                        return False
                    for (ac, astms), (bc, bstms) in zip(self.elIfs,
                                                        other.elIfs):
                        if not (ac == bc) or\
                                not isSameStatementList(astms, bstms):
                            return False
                    return True
        return False

    def seqEval(self):
        if seqEvalCond(self.cond):
            for s in self.ifTrue:
                s.seqEval()
        else:
            for c in self.elIfs:
                if seqEvalCond(c[0]):
                    for s in c[1]:
                        s.seqEval()
                    return

            for s in self.ifFalse:
                s.seqEval()


class SwitchContainer(HdlObject):
    """
    Structural container for switch statement for hdl rendering
    """

    def __init__(self, switchOn, cases, default=[]):
        self.switchOn = switchOn
        self.cases = cases
        self.default = default

    def isSame(self, other):
        if isinstance(other, SwitchContainer) \
                and isSameHVal(self.switchOn, other.switchOn)\
                and len(self.cases) == len(other.cases)\
                and isSameStatementList(self.default, other.default):
            for (ac, astm), (bc, bstm) in zip(self.cases, other.cases):
                if not isSameHVal(ac, bc)\
                        or not isSameStatementList(astm, bstm):
                    return False
            return True
        return False

    def seqEval(self):
        raise NotImplementedError()


class WhileContainer(HdlObject):
    """
    Structural container of while statement for hdl rendering
    """

    def __init__(self, cond, body):
        self.cond = cond
        self.body = body

    def seqEval(self):
        while seqEvalCond(self.cond):
            for s in self.body:
                s.seqEval()


class WaitStm(HdlObject):
    """
    Structural container of wait statemnet for hdl rendering
    """

    def __init__(self, waitForWhat):
        self.isTimeWait = isinstance(waitForWhat, int)
        self.waitForWhat = waitForWhat



from hwt.hdl.variables import SignalItem
from hwt.serializer.generic.value import GenericSerializer_Value
from hwt.serializer.generic.indent import getIndent
from hwt.synthesizer.param import Param, evalParam


class HwtSerializer_value(GenericSerializer_Value):

    @classmethod
    def Bits_valAsHdl(cls, dtype, val, ctx):
        return "%s.fromPy(%d, vldMask=%d)" % (
            cls.HdlType_bits(dtype, ctx, declaration=False),
            val.val, val.vldMask)

    @classmethod
    def RtlSignal(cls, s, ctx, declaration=False):
        return cls.SignalItem(s, ctx, declaration=declaration)

    @classmethod
    def SignalItem(cls, si, ctx, declaration=False):
        if declaration:
            raise NotImplementedError()
        else:
            if isinstance(si, Param):
                return cls.Value(evalParam(si), ctx)
            if isinstance(si, SignalItem) and si._const:
                return cls.Value(si._val, ctx)
            if si.hidden and hasattr(si, "origin"):
                return cls.asHdl(si.origin, ctx)
            else:
                return "%s" % si.name

    @classmethod
    def Integer_valAsHdl(cls, t, i, ctx):
        if i.vldMask:
            return "%d" % i.val
        else:
            return "None"

    @classmethod
    def Dict_valAsHdl(cls, val, ctx):
        sep = (",\n" + getIndent(ctx.indent + 1))

        def sItem(i):
            k, v = i
            return "%d: %s" % (k, cls.Value(v, ctx))

        return "{%s}" % sep.join(map(sItem, val.items()))

    @classmethod
    def HArrayValAsHdl(cls, t, val, ctx):
        return "HArrayVal(%s, %s, %d)" % (
            cls.Dict_valAsHdl(val.val, ctx),
            cls.HdlType(t, ctx),
            val.vldMask)

    @classmethod
    def Slice_valAsHdl(cls, t, val, ctx):
        return "SliceVal((simHInt(%d), simHInt(%d)), SLICE, %d)" % (
            evalParam(val.val[0]).val,
            evalParam(val.val[1]).val,
            val.vldMask)

    @classmethod
    def HEnumValAsHdl(cls, t, val, ctx):
        return "%s.%s" % (t.name, val.val)

    @classmethod
    def condAsHdl(cls, cond, ctx):
        cond = list(cond)
        return "%s" % (",".join(map(lambda x: cls.asHdl(x, ctx), cond)))



from typing import List

from hwt.hdl.architecture import Architecture
from hwt.hdl.assignment import Assignment
from hwt.hdl.entity import Entity
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.process import HWProcess
from hwt.hdl.statements import IfContainer, SwitchContainer
from hwt.hdl.types.array import HArray
from hwt.hdl.value import Value
from hwt.serializer.generic.serializer import GenericSerializer
from hwt.serializer.resourceAnalyzer.resourceTypes import Unconnected, \
    ResourceMUX, ResourceLatch, ResourceLatchWithMux
from hwt.serializer.resourceAnalyzer.utils import ResourceContext, \
    updateGuesFromAssignment, resourceTransitions_override, mergeGues, \
    operatorsWithoutResource
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class ResourceAnalyzer(GenericSerializer):
    """
    Serializer which does not products any output just collect informations
    about used resources

    :attention: Use instance of ResourceAnalyzer instead of class
    """
    _keywords_dict = {}

    def __init__(self):
        self.context = ResourceContext(None)

    @classmethod
    def HWProcess(cls, proc: HWProcess, ctx: ResourceContext) -> None:
        """
        Gues resource usage by HWProcess
        """
        gues = cls.statementList(proc.statements, proc.outputs, ctx)
        for sig, resGues in gues.items():
            ctx.register(sig, resGues)

    @classmethod
    def statementList(cls, statements,
                      dstSignals,
                      ctx: ResourceContext) -> None:
        """
        Gues resource usage for list of statements

        :param dstSignals: list of signals which are driven by this assignments 
        """
        gues = {}
        guesOfChildren = []
        # resolve
        onlyAssignmets = True
        for stm in statements:
            if isinstance(stm, Assignment):
                updateGuesFromAssignment(gues, stm)
                if isinstance(stm.src, RtlSignal):
                    cls.Signal(stm.src, ctx)
            else:
                onlyAssignmets = False
                fn = getattr(cls, stm.__class__.__name__)
                g = fn(stm, dstSignals, ctx)
                guesOfChildren.append(g)

        for g in guesOfChildren:
            for sig, childGues in g.items():
                try:
                    current = gues[sig]
                except KeyError:
                    current = Unconnected

                nextGues = resourceTransitions_override[(current, childGues)]
                gues[sig] = nextGues

        # mark signals which were not connected as unconnected
        if onlyAssignmets and len(gues) != len(dstSignals):
            for sig in dstSignals:
                if sig not in gues.keys():
                    gues[sig] = Unconnected

        return gues

    @classmethod
    def condition(cls, condition: List[RtlSignal],
                  ctx: ResourceContext) -> None:
        """
        Gues resource usage by this condition expression
        """
        for signal in condition:
            cls.Signal(signal, ctx)

    @classmethod
    def IfContainer(cls, ifc: IfContainer,
                    dstSignals,
                    ctx: ResourceContext):
        """
        Gues resource usage by this if statement

        :param dstSignals: list of signals which are driven by this assignments
        """
        cls.condition(ifc.cond, ctx)

        ifTrue = cls.statementList(ifc.ifTrue, dstSignals, ctx)

        for elifCond, elifStm in ifc.elIfs:
            cls.condition(elifCond, ctx)
            elifGues = cls.statementList(elifStm, dstSignals, ctx)
            mergeGues(ifTrue, elifGues)

        if ifc.ifFalse:
            ifFalse = cls.statementList(ifc.ifFalse, dstSignals, ctx)
        else:
            ifFalse = {s: Unconnected for s in dstSignals}
        mergeGues(ifTrue, ifFalse)

        return ifTrue

    @classmethod
    def SwitchContainer(cls, swc: SwitchContainer,
                        dstSignals,
                        ctx: ResourceContext) -> None:
        """
        Gues resource usage by this switch statement

        :param dstSignals: list of signals which are driven by this assignments
        """
        isEnclosed = swc.default or len(swc.cases) >= (
            1 << swc.switchOn._dtype.bit_length())
        gues = None
        for k, c in swc.cases:
            g = cls.statementList(c, dstSignals, ctx)
            if gues is None:
                gues = g
            else:
                mergeGues(gues, g)

        if swc.default:
            g = cls.statementList(swc.default, dstSignals, ctx)
            if gues is None:
                gues = g
            else:
                mergeGues(gues, g)

        if not isEnclosed:
            # convert all multiplexers or assignments to latches
            update = {}
            for k, v in gues.items():
                if v is Assignment:
                    update[k] = ResourceLatch
                elif v is ResourceMUX:
                    update[k] = ResourceLatchWithMux
            gues.update(update)

        return gues

    @classmethod
    def operator(cls, driver: Operator, signal: RtlSignal,
                 ctx: ResourceContext):
        """
        Resolve resource usage by operator
        """
        op = driver.operator
        if op is AllOps.TERNARY:
            # ternary can be to Bits(1) conversion
            try:
                a = bool(driver.operands[1])
                b = bool(driver.operands[2])
            except ValueError:
                a = False
                b = False

            if a and not b:
                # this is just to bit conversion if cond: 1 else 0
                pass
            else:
                # this is multiplexer
                width = signal._dtype.bit_length()
                inputsCnt = 2
                ctx.registerMUX_known(width, inputsCnt)

        else:
            doRegister = True
            # check to bool conversion
            if (op is AllOps.EQ
                    and driver.operands[0]._dtype.bit_length() == 1):
                try:
                    a = bool(driver.operands[1])
                    # if this is just to bool conversion skip it
                    doRegister = not a
                except ValueError:
                    doRegister = True

            if doRegister:
                ctx.registerOperator(driver)

    @classmethod
    def Signal(cls, signal: RtlSignal, ctx: ResourceContext) -> None:
        """
        Resolve resources for signal by walking it's drivers
        """
        # skip constants because they does not consume any resources
        # directly
        if not signal._const and signal not in ctx.seen:
            ctx.seen.add(signal)
            # walk drivers to find rest of resources in expression
            for driver in signal.drivers:
                if isinstance(driver, Operator):
                    op = driver.operator
                    if op is AllOps.INDEX:
                        # index can be MUX/memory/static bit select
                        i = driver.operands[1]
                        if isinstance(i, Value) or i._const:
                            # bit selecting operator
                            pass
                        else:
                            # mux or read ram port
                            inputsSig = driver.operands[0]
                            if isinstance(inputsSig._dtype, HArray):
                                # this indexing is read port of some RAM/ROM
                                ctx.discoveredRamSignals.add(inputsSig)
                            else:
                                # this indexing is MUX
                                width = driver.result._dtype.bit_length()
                                inputs = inputsSig._dtype.bit_length() // width
                                ctx.registerMUX_known(width, inputs)
                    # skip conversions/clk ops etc. which does not consume
                    # resources directly
                    elif op not in operatorsWithoutResource:
                        cls.operator(driver, signal, ctx)

                    # collect resources for other operators as well
                    for oper in driver.operands:
                        if isinstance(oper, RtlSignal):
                            cls.Signal(oper, ctx)

                elif isinstance(driver, Assignment):
                    # processed by HWProcess
                    pass
                else:
                    raise NotImplementedError(driver)

    @classmethod
    def Entity(cls, ent: Entity, ctx: ResourceContext) -> None:
        """
        Entity is just header, we do not need to inspect it for resources
        """
        return

    def getBaseContext(self) -> ResourceContext:
        """
        Return context for collecting of resource informatins
        prepared on this instance
        """
        return self.context

    @classmethod
    def Architecture(cls, arch: Architecture, ctx: ResourceContext) -> None:
        for c in arch.componentInstances:
            raise NotImplementedError()

        for proc in arch.processes:
            cls.HWProcess(proc, ctx)

        ctx.finalize()

    def report(self):
        return self.context.resources



from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.value import Value
from hwt.hdl.variables import SignalItem
from hwt.simulator.exceptions import SimException
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.synthesizer.rtlLevel.signalUtils.ops import RtlSignalOps
from hwt.synthesizer.uniqList import UniqList


class RtlSignal(RtlSignalBase, SignalItem, RtlSignalOps):
    """
    RtlSignal signal is container of connection between statements and operators

    :ivar _usedOps: dictionary of used operators which can be reused
    :ivar endpoints: UniqList of operators and statements
        for which this signal is driver.
    :ivar drivers: UniqList of operators and statements
        which can drive this signal.
    :ivar hiden: means that this signal is part of expression
        and should not be rendered
    :ivar processCrossing: means that this signal is crossing process boundary

    :cvar __instCntr: counter used for generating instance ids
    :ivar _instId: internally used only for intuitive sorting of statements
        in serialized code
    """
    __instCntr = 0

    def __init__(self, ctx, name, dtype, defaultVal=None, nopVal=None,
                 useNopVal=False):
        """
        :param ctx: context - RtlNetlist which is this signal part of
        :param name: name hint for this signal, if is None name
            is chosen automatically
        :param defaultVal: value which is used for reset and as default value
            in hdl
        :param useNopVal: use nopVal or ignore it
        :param nopVal: value which is used to fill up statements when no other
            value is assigned
        """

        if name is None:
            name = "sig_"
            self.hasGenericName = True
        else:
            self.hasGenericName = False

        assert isinstance(dtype, HdlType)
        super(RtlSignal, self).__init__(name, dtype, defaultVal)
        self.ctx = ctx

        if ctx:
            # params does not have any context on created
            # and it is assigned after param is bounded to unit or interface
            ctx.signals.add(self)

        # set can not be used because hash of items are changing
        self.endpoints = UniqList()
        self.drivers = UniqList()
        self._usedOps = {}
        self.hidden = True
        self._instId = RtlSignal._nextInstId()

        self._nopVal = nopVal
        self._useNopVal = useNopVal
        self._const = False

    @classmethod
    def _nextInstId(cls):
        """
        Get next instance id
        """
        i = cls.__instCntr
        cls.__instCntr += 1
        return i

    def staticEval(self):
        # operator writes in self._val new value
        if self.drivers:
            for d in self.drivers:
                d.staticEval()
        else:
            if isinstance(self.defaultVal, RtlSignal):
                self._val = self.defaultVal._val.staticEval()
            else:
                if self._val.updateTime < 0:
                    self._val = self.defaultVal.clone()

        if not isinstance(self._val, Value):
            raise SimException(
                "Evaluation of signal returned not supported object (%r)"
                % (self._val, ))
        return self._val

    def singleDriver(self):
        """
        Returns a first driver if signal has only one driver.
        """
        if len(self.drivers) != 1:
            raise MultipleDriversExc()
        return list(self.drivers)[0]



from hwt.hdl.variables import SignalItem
from hwt.hdl.constants import DIRECTION


class PortItem(SignalItem):
    """basic hdl entity port item"""

    def __init__(self, name, direction, dtype, unit):
        self.name = name
        self.unit = unit
        self.direction = direction
        self._dtype = dtype
        self.src = None
        self.dst = None

    def connectSig(self, signal):
        """
        Connect to port item on subunit
        """
        if self.direction == DIRECTION.IN:
            if self.src is not None:
                raise Exception("Port %s is already associated with %s" % (
                    self.name, str(self.src)))
            self.src = signal

            signal.endpoints.append(self)
        elif self.direction == DIRECTION.OUT:
            if self.dst is not None:
                raise Exception("Port %s is already associated with %s" % (
                    self.name, str(self.dst)))
            self.dst = signal

            signal.drivers.append(self)
        else:
            raise NotImplementedError()

        signal.hidden = False
        signal.ctx.subUnits.add(self.unit)

    def reigsterInternSig(self, signal):
        """
        Connect internal signal to port item,
        this connection is used by simulator and only output port items
        will be connected
        """
        if self.direction == DIRECTION.OUT:
            if self.src is not None:
                raise Exception("Port %s is already associated with %s" % (
                    self.name, str(self.src)))
            self.src = signal
        elif self.direction == DIRECTION.IN:
            if self.dst is not None:
                raise Exception("Port %s is already associated with %s" % (
                    self.name, str(self.dst)))
            self.dst = signal
        else:
            raise NotImplementedError()

    def connectInternSig(self):
        """
        connet signal from internal side of of this component to this port
        """
        if self.direction == DIRECTION.OUT:
            self.src.endpoints.append(self)
        elif self.direction == DIRECTION.IN:
            self.dst.drivers.append(self)
        elif self.direction == DIRECTION.INOUT:
            self.dst.drivers.append(self)
        else:
            raise NotImplementedError()

    def getSigInside(self):
        """
        return signal inside unit which has this port
        """
        d = self.direction
        if d is DIRECTION.IN:
            return self.dst
        elif d is DIRECTION.OUT:
            return self.src
        else:
            raise NotImplementedError(d)



import keyword


SIMMODEL_KEYWORDS = keyword.kwlist + [
    # names defined in SimModel constructor
    '_cntx',
    '_interfaces',
    '_processes',
    '_units',
    '_outputs',
    # imports
    'hwt',
    'importlib',
    'reload',
    'HArray',
    'HArrayVal',
    'convertBits__val',
    'BitsVal',
    'SLICE',
    'HEnum'
    'DIRECTION',
    'SENSITIVITY',
    'convertSimInteger__val',
    'simHInt',
    'SIM_INT',
    'simBitsT',
    'SIM_BIT',
    'convertSimBits__val',
    'SimModel',
    'sensitivity',
    'connectSimPort',
    'simEvalCond',
    'mkUpdater',
    'mkArrayUpdater'
    'power'
    'RtlNetlist'
    'SimSignal'
    'SliceVal']



from jinja2.environment import Environment
from jinja2.loaders import PackageLoader

from hwt.hdl.architecture import Architecture
from hwt.hdl.assignment import Assignment
from hwt.hdl.constants import SENSITIVITY, DIRECTION
from hwt.hdl.entity import Entity
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.process import HWProcess
from hwt.hdl.statements import IfContainer, SwitchContainer
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.enumVal import HEnumVal
from hwt.serializer.generic.serializer import GenericSerializer
from hwt.serializer.hwt.keywords import HWT_KEYWORDS
from hwt.serializer.hwt.ops import HwtSerializer_ops
from hwt.serializer.hwt.types import HwtSerializer_types
from hwt.serializer.hwt.value import HwtSerializer_value
from hwt.serializer.generic.constCache import ConstCache
from hwt.serializer.generic.context import SerializerCtx
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.generic.nameScope import LangueKeyword
from hwt.serializer.utils import maxStmId
from hwt.synthesizer.param import evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdl.value import Value
from hwt.serializer.exceptions import SerializerException


env = Environment(loader=PackageLoader('hwt', 'serializer/hwt/templates'))
unitHeadTmpl = env.get_template('unit_head.py.template')
unitBodyTmpl = env.get_template('unit_body.py.template')
processTmpl = env.get_template('process.py.template')
ifTmpl = env.get_template("if.py.template")


class HwtSerializer(HwtSerializer_value, HwtSerializer_ops,
                    HwtSerializer_types, GenericSerializer):
    """
    Serializer which converts Hwt objects back to Hwt code
    for debugging purposes
    """
    _keywords_dict = {kw: LangueKeyword() for kw in HWT_KEYWORDS}
    fileExtension = '.py'

    @classmethod
    def serializationDecision(cls, obj, serializedClasses,
                              serializedConfiguredUnits):
        return True

    @classmethod
    def asHdl(cls, obj, ctx):
        """
        Convert object to HDL string

        :param obj: object to serialize
        :param ctx: SerializerCtx instance
        """
        if isinstance(obj, RtlSignalBase):
            return cls.SignalItem(obj, ctx)
        elif isinstance(obj, Value):
            return cls.Value(obj, ctx)
        else:
            try:
                serFn = obj.asHwt
            except AttributeError:
                serFn = None
            if serFn is not None:
                return serFn(cls, ctx)

            try:
                serFn = getattr(cls, obj.__class__.__name__)
            except AttributeError:
                serFn = None

            if serFn is not None:
                return serFn(obj, ctx)

            raise SerializerException("Not implemented for %r" % (obj))

    @classmethod
    def stmAsHdl(cls, obj, ctx: SerializerCtx, enclosure=None):
        try:
            serFn = getattr(cls, obj.__class__.__name__)
        except AttributeError:
            raise NotImplementedError("Not implemented for %s" % (repr(obj)))
        return serFn(obj, ctx, enclosure=enclosure)

    @classmethod
    def Entity(cls, ent: Entity, ctx: SerializerCtx):
        """
        Entity is just forward declaration of Architecture, it is not used
        in most HDL languages as there is no recursion in hierarchy
        """

        cls.Entity_prepare(ent, ctx, serialize=False)
        ent.name = ctx.scope.checkedName(ent.name, ent, isGlobal=True)
        ports = list(
            map(lambda p: (p.name, cls.HdlType(p._dtype, ctx)),
                ent.ports))
        return unitHeadTmpl.render(
            name=ent.name,
            ports=ports,
        )

    @classmethod
    def Architecture(cls, arch: Architecture, ctx: SerializerCtx):
        variables = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: x.name)
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        arch.componentInstances.sort(key=lambda x: x._name)

        for v in arch.variables:
            t = v._dtype
            # if type requires extra definition
            if isinstance(t, HEnum) and t not in extraTypes:
                extraTypes.add(v._dtype)
                extraTypes_serialized.append(
                    cls.HdlType(t, ctx, declaration=True))

            v.name = ctx.scope.checkedName(v.name, v)
            variables.append(v)

        childCtx = ctx.withIndent(2)
        childCtx.constCache = ConstCache(ctx.scope.checkedName)

        def serializeVar(v):
            dv = evalParam(v.defaultVal)
            if isinstance(dv, HEnumVal):
                dv = "%s.%s" % (dv._dtype.name, dv.val)
            else:
                dv = cls.Value(dv, ctx)

            return v.name, cls.HdlType(v._dtype, childCtx), dv

        for p in arch.processes:
            procs.append(cls.HWProcess(p, childCtx))

        constants = []
        for c in sorted(childCtx.constCache._cache.items(), key=lambda x: x[1],
                        reverse=True):
            constants.append((c[1], cls.Value(c[0], ctx)))

        portNames = [p.name for p in arch.entity.ports]
        portToLocalsRow = "%s = %s" % (
            ", ".join(portNames),
            ", ".join(["self." + n for n in portNames]))

        return unitBodyTmpl.render(
            DIRECTION_IN=DIRECTION.IN,
            name=arch.getEntityName(),
            portToLocalsRow=portToLocalsRow,
            constants=constants,
            signals=list(map(serializeVar, variables)),
            extraTypes=extraTypes_serialized,
            processes=procs,
            componentInstances=arch.componentInstances,
        )

    @classmethod
    def Assignment(cls, a: Assignment, ctx: SerializerCtx, enclosure=None):
        dst = a.dst
        indentStr = getIndent(ctx.indent)
        srcStr = "%s" % cls.Value(a.src, ctx)
        indexes = a.indexes
        if indexes is None:
            indexes = []

        return "%s%s%s(%s)" % (
            indentStr, dst.name,
            "".join(map(lambda x: "[%s]" % cls.asHdl(x, ctx),
                        indexes)),
            srcStr
        )

    @classmethod
    def comment(cls, coment: str):
        return "#" + coment.replace("\n", "\n#")

    @classmethod
    def IfContainer(cls, ifc: IfContainer, ctx: SerializerCtx, enclosure=None):
        cond = cls.condAsHdl(ifc.cond, ctx)
        ifTrue = ifc.ifTrue
        ifFalse = ifc.ifFalse
        childCtx = ctx.withIndent()
        if enclosure is None:
            _enclosure = "# no enclosure"
        else:
            _enclosure = "\n".join(
                [cls.stmAsHdl(e, childCtx) for e in enclosure])

        def serialize_statements(statements):
            return [cls.stmAsHdl(obj, childCtx, enclosure=enclosure)
                    for obj in statements]

        def serialize_elif(elifCase):
            cond, statements = elifCase
            return (cls.condAsHdl(cond, ctx),
                    serialize_statements(statements))

        return ifTmpl.render(
            indent=getIndent(ctx.indent),
            indentNum=ctx.indent,
            cond=cond,
            enclosure=_enclosure,
            ifTrue=serialize_statements(ifTrue),
            elIfs=[serialize_elif(elIf) for elIf in ifc.elIfs],
            ifFalse=serialize_statements(ifFalse)
        )

    @classmethod
    def SwitchContainer(cls, sw: SwitchContainer,
                        ctx: SerializerCtx, enclosure=None):
        switchOn = sw.switchOn

        def mkCond(c):
            return {Operator(AllOps.EQ,
                             [switchOn, c])}
        elIfs = []

        for key, statements in sw.cases:
            elIfs.append((mkCond(key), statements))
        ifFalse = sw.default

        topCond = mkCond(sw.cases[0][0])
        topIf = IfContainer(topCond,
                            sw.cases[0][1],
                            ifFalse,
                            elIfs)

        return cls.IfContainer(topIf, ctx, enclosure=enclosure)

    @classmethod
    def sensitivityListItem(cls, item):
        if isinstance(item, Operator):
            op = item.operator
            if op == AllOps.RISING_EDGE:
                sens = SENSITIVITY.RISING
            elif op == AllOps.FALLIGN_EDGE:
                sens = SENSITIVITY.FALLING
            else:
                raise TypeError("This is not an event sensitivity", op)

            return "(%s, %s)" % (str(sens), item.operands[0].name)
        else:
            return item.name

    @classmethod
    def HWProcess(cls, proc: HWProcess, ctx: SerializerCtx):
        body = proc.statements
        proc.name = ctx.scope.checkedName(proc.name, proc)
        sensitivityList = sorted(
            map(cls.sensitivityListItem, proc.sensitivityList))

        if len(body) == 1:
            _body = cls.stmAsHdl(body[0], ctx)
        else:
            for i, stm in enumerate(body):
                if not isinstance(stm, Assignment):
                    break
            # first statement is taken as default
            enclosure = body[:i]
            _body = "\n".join([cls.stmAsHdl(stm, ctx, enclosure=enclosure)
                               for stm in body[i:]])

        return processTmpl.render(
            indent=getIndent(ctx.indent),
            name=proc.name,
            sensitivityList=sensitivityList,
            stmLines=[_body]
        )



from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.defs import BOOL
from hwt.hdl.value import Value

BoolVal = BOOL.getValueCls()


class EventCapableVal(Value):
    """
    Base class for event capable values
    """

    def _onFallingEdge__val(self, now):
        v = BoolVal(self.updateTime == now,
                    BOOL,
                    self.vldMask,
                    now)
        v.val = v.val and not self.val
        return v

    def _onFallingEdge(self, now):
        if isinstance(self, Value):
            return self._onFallingEdge__val(now)
        else:
            return Operator.withRes(AllOps.FALLIGN_EDGE, [self], BOOL)

    def _onRisingEdge__val(self, now):
        v = BoolVal(self.updateTime == now,
                    BOOL,
                    self.vldMask,
                    now)
        v.val = v.val and self.val
        return v

    def _onRisingEdge(self, now):
        if isinstance(self, Value):
            return self._onRisingEdge__val(now)
        else:
            return Operator.withRes(AllOps.RISING_EDGE, [self], BOOL)



from hwt.hdl.constants import DIRECTION
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.struct import HStruct, HStructField, HStructFieldMeta
from hwt.interfaces.agents.structIntf import StructIntfAgent
from hwt.interfaces.std import Signal, VldSynced, RegCntrl, BramPort_withoutClk
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.interfaceLevel.unitImplHelpers import getSignalName
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class StructIntf(Interface):
    """
    Create dynamic interface based on HStruct or HUnion description

    :ivar _fieldsToInterfaces: dictionary {field from HStruct template:
        sub interface for it}
    :ivar _structT: HStruct instance used as template for this interface
    :param _instantiateFieldFn: function(FieldTemplateItem instance)
        return interface instance
    """

    def __init__(self, structT, instantiateFieldFn,
                 masterDir=DIRECTION.OUT, asArraySize=None,
                 loadConfig=True):
        Interface.__init__(self,
                           masterDir=masterDir,
                           asArraySize=asArraySize,
                           loadConfig=loadConfig)
        self._structT = structT
        self._instantiateFieldFn = instantiateFieldFn
        self._fieldsToInterfaces = {}

    def _declr(self):
        _t = self._structT
        if isinstance(_t, HStruct):
            fields = _t.fields
        else:
            fields = _t.fields.values()

        self._fieldsToInterfaces[self._structT] = self

        for field in fields:
            # skip padding
            if field.name is not None:
                # generate interface based on struct field
                intf = self._instantiateFieldFn(self, field)
                self._fieldsToInterfaces[field] = intf
                setattr(self, field.name, intf)

                if isinstance(intf, StructIntf):
                    intf._fieldsToInterfaces = self._fieldsToInterfaces

    def _initSimAgent(self):
        self._ag = StructIntfAgent(self)


def _HTypeFromIntfMap(intf):
    name = getSignalName(intf)
    if isinstance(intf, (RtlSignalBase, Signal)):
        dtype = intf._dtype
    elif isinstance(intf, VldSynced):
        dtype = intf.data._dtype
    elif isinstance(intf, RegCntrl):
        dtype = intf.din._dtype
    elif isinstance(intf, BramPort_withoutClk):
        dtype = Bits(int(intf.DATA_WIDTH))[2 ** int(intf.ADDR_WIDTH)]
    else:
        dtype, name = intf
        assert isinstance(dtype, HdlType)
        assert isinstance(name, str)

    return (dtype, name)


def isIntfMap(intfMap):
    if not isinstance(intfMap, tuple):
        return False
    elif len(intfMap) == 2 and (intfMap[1] is None
                                or isinstance(intfMap, str)):
        return False
    else:
        return True


def HTypeFromIntfMapItem(interfaceMapItem):
    isTerminal = False
    if isinstance(interfaceMapItem, (InterfaceBase, RtlSignalBase)):
        dtype, nameOrPrefix = _HTypeFromIntfMap(interfaceMapItem)
        isTerminal = True
    else:
        typeOrListOfInterfaces, nameOrPrefix = interfaceMapItem

        if isinstance(typeOrListOfInterfaces, list):
            # tuple (list or items, name of this array)
            types = []
            reference = None

            for item in typeOrListOfInterfaces:
                if isIntfMap(item):
                    t = HTypeFromIntfMap(item)
                else:
                    f = HTypeFromIntfMapItem(item)
                    t = f.dtype

                types.append(t)
                if reference is None:
                    reference = t
                else:
                    assert reference == t, (
                        "all items in array has to have same type")

            dtype = reference[len(types)]

        elif isinstance(typeOrListOfInterfaces, HdlType):
            dtype = typeOrListOfInterfaces
            isTerminal = True
        elif isinstance(typeOrListOfInterfaces,
                        (InterfaceBase, RtlSignalBase)):
            # renamed interface, ignore original name
            dtype = _HTypeFromIntfMap(typeOrListOfInterfaces)[0]
            isTerminal = True

        else:
            # tuple (tuple of interfaces, prefix)
            assert isinstance(typeOrListOfInterfaces,
                              tuple), typeOrListOfInterfaces
            dtype = HTypeFromIntfMap(typeOrListOfInterfaces)

    assert isinstance(nameOrPrefix, str) or nameOrPrefix is None, nameOrPrefix

    f = HStructField(dtype, nameOrPrefix)

    if not isTerminal:
        f.meta = HStructFieldMeta(split=True)

    return f


def HTypeFromIntfMap(interfaceMap):
    """
    Generate flattened register map for HStruct

    :param interfaceMap: sequence of
        tuple (type, name) or (will create standard struct field member)
        interface or (will create a struct field from interface)
        instance of hdl type (is used as padding)
        tuple (list of interface, name)
    :param DATA_WIDTH: width of word
    :param terminalNodes: None or set whre are placed StructField instances
        which are derived directly from interface
    :return: generator of tuple (type, name, BusFieldInfo)
    """
    structFields = []

    for m in interfaceMap:
        f = HTypeFromIntfMapItem(m)
        structFields.append(f)

    return HStruct(*structFields)
